#include	<stdlib.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#include	<yz/image.h>

#define		iabs(x)	((x>0)?x:(-x))

#define	YZ_UITK_IMPLEMENTOR
#include	"uitk.h"

#include	"scrollbar.h"

struct scrollarea_t
{
	int	type;
	int	x, y, dx, dy;	// on-window position
	int	ox, oy;			// current origin
	int	in_drag;
	void	(*update)(void	*data, int	x1, int	y1, int	x2, int	y2);
	void	(*event)(void	*data, event_t	ev);
	void	*data;
};

static int	handle_mouse_event_in_scroll_area(scrollarea_t	*a, event_t	*ev);

#define	TYPE_NAME	"scroll-area"

static int						this_type_filter(event_t	*ev);
static int						thisTypeNum = 0;
static ui_element_type_info		typeInfo = { TYPE_NAME, 0, 0, this_type_filter };

static int register_this_type()
{
	thisTypeNum = yzUIRegisterTypeInfo(&typeInfo);
	if(thisTypeNum==0)return -1;
	return 0;
}

#define	CHECK_REG()	if(register_this_type())return 0;

scrollarea_t	*yzAddScrollArea(int	x, int	y, int	dx, int	dy,
				void	(*update)(void	*data, int	x1, int	y1, int	x2, int	y2),
				void	(*event)(void	*data, event_t	ev), void	*data)
{
	scrollarea_t	*a;
	
	CHECK_REG();
	a = (scrollarea_t*)malloc(sizeof(scrollarea_t));
	a->type = thisTypeNum;
	a->x = x; a->y = y;
	a->dx = dx; a->dy = dy;
	a->ox = 0; a->oy = 0;
	a->in_drag = 0;
	a->data = data;
	a->update = update;
	a->event = event;
	yzUIAddToList(a);
	return a;
}

void	yzResizeScrollArea(scrollarea_t	*a, int	x, int	y, int	dx, int	dy)
{
	a->x = x;
	a->y = y;
	a->dx = dx;
	a->dy = dy;
}

void	yzRemoveScrollArea(scrollarea_t	*a)
{
	if(!a)return;
	yzUIRemoveFromList(a);
	free((void*)a);
}

void		yzResetScrollArea(scrollarea_t	*a)
{
	a->ox = a->oy = 0;
}

void	yzRecomputeScrollAreaClipping(scrollarea_t	*a)
{
	yzPushClipRect(a->x, a->y, a->x+a->dx-1, a->y+a->dy-1);
	yzPushOrigin(a->x + a->ox, a->y + a->oy);
}

void	yzDrawScrollArea(scrollarea_t	*a)
{
	yzPushClipRect(a->x, a->y, a->x+a->dx-1, a->y+a->dy-1);
	yzBufferMode(YZ_BUFFER_BACK);
	yzPushOrigin(a->x + a->ox, a->y + a->oy);
	a->update(a->data, -a->ox, -a->oy, -a->ox+a->dx-1, -a->oy+a->dy-1);
	yzPopOrigin();
	yzUpdateBuffer(YZ_BUFFER_FRONT);
	yzBufferMode(YZ_BUFFER_BOTH);
	yzPopClipRect();
}

void	yzUpdateScrollArea(scrollarea_t	*a, int	x1, int	y1, int	x2, int	y2)
{
	int	vx1, vy1, vx2, vy2;

	vx1 = -a->ox;
	vx2 = -a->ox + a->dx - 1;
	vy1 = -a->oy;
	vy2 = -a->oy + a->dy - 1;
	if(x1<vx1)x1=vx1;
	if(x2>vx2)x2=vx2;
	if(y1<vy1)y1=vy1;
	if(y2>vy2)y2=vy2;
	if(x1>x2 || y1>y2)return;	// not onscreen
	yzBufferMode(YZ_BUFFER_BACK);
	yzPushClipRect(a->x, a->y, a->x+a->dx-1, a->y+a->dy-1);
	yzPushOrigin(a->x + a->ox, a->y + a->oy);
	a->update(a->data, x1, y1, x2, y2);
	yzPopOrigin();
	yzUpdateBuffer(YZ_BUFFER_FRONT);
	yzBufferMode(YZ_BUFFER_BOTH);
	yzPopClipRect();
}

void	effect_vertical_scroll(scrollarea_t	*a, int	dy)
{
	int	y1, y2;

	if(!dy)return;

	a->oy += dy;

	// if we're moving more than a screenful, just redraw the whole thing
	if(dy>=a->dy || dy<=-a->dy)
	{
		yzDrawScrollArea(a);
		return;
	}

	// clip
	yzPushClipRect(a->x, a->y, a->x+a->dx-1, a->y+a->dy-1);
	yzBufferMode(YZ_BUFFER_BACK);

	if(dy>0)
		yzCopyRect(a->x, a->y, a->x, a->y + dy, a->dx, a->dy - dy);
	else
		yzCopyRect(a->x, a->y-dy, a->x, a->y, a->dx, a->dy + dy);

	if(dy>0){y1 = 0;y2 = dy-1;}
	else	{y1 = a->dy-1+(dy-1);y2 = a->dy-1;}
	yzPushClipRect(a->x, a->y + y1, a->x+a->dx-1, a->y + y2);
	yzPushOrigin(a->x + a->ox, a->y + a->oy);
	a->update(a->data, -a->ox, -a->oy+y1, -a->ox+a->dx-1, -a->oy+y2);
	yzPopOrigin();
	yzPopClipRect();

	yzUpdateBuffer(YZ_BUFFER_FRONT);
	yzBufferMode(YZ_BUFFER_BOTH);

	yzPopClipRect();
}

void	effect_horizontal_scroll(scrollarea_t	*a, int	dx)
{
	int	x1, x2;
	
	if(!dx)return;

	a->ox += dx;

	// if we're moving more than a screenful, just redraw the whole thing
	if(dx>=a->dx || dx<=-a->dx)
	{
		yzDrawScrollArea(a);
		return;
	}

	// clip to the scroll area
	yzPushClipRect(a->x, a->y, a->x+a->dx-1, a->y+a->dy-1);
	// do work in back buffer
	yzBufferMode(YZ_BUFFER_BACK);

	// move stuff if possible
	if(dx>0)
		yzCopyRect(a->x, a->y, a->x + dx, a->y, a->dx - dx, a->dy);
	else
		yzCopyRect(a->x - dx, a->y, a->x, a->y, a->dx + dx, a->dy);

	// draw ONLY the area exposed
	if(dx>0){x1 = 0;x2 = dx-1;}
	else	{x1 = a->dx-1+(dx-1);x2 = a->dx-1;}
	yzPushClipRect(a->x + x1, a->y, a->x + x2, a->y+a->dy-1);
	yzPushOrigin(a->x + a->ox, a->y + a->oy);
	a->update(a->data, -a->ox+x1, -a->oy, -a->ox + x2, -a->oy+a->dy-1);
	yzPopOrigin();
	yzPopClipRect();

	// copy the scroll area to front buffer
	
	yzUpdateBuffer(YZ_BUFFER_FRONT);
	yzBufferMode(YZ_BUFFER_BOTH);

	yzPopClipRect();
}

void	effect_scroll(scrollarea_t	*a, int	dx, int	dy)
{
	if(dx==0)effect_vertical_scroll(a, dy);
	else if(dy==0)effect_horizontal_scroll(a, dx);
}

static int	point_in_coords(int	x, int	y, int	x0, int	y0, int	x1, int	y1)
{
	if(x>=x0 && x<=x1 && y>=y0 && y<=y1)return 1;
	return 0;
}

static int		this_type_filter(event_t	*ev)
{
	ui_element_list	*l;
	scrollarea_t		*a;
	
	l = yzUIGetList(TYPE_NAME);
	if(!l)return 0;
	switch(ev->type)
	{
		case	YZ_MOUSE_DOWN:
			while(l)
			{
				a = l->e;
				if(point_in_coords(ev->x, ev->y, a->x, a->y,
					a->x+a->dx-1, a->y+a->dy-1))
				{
					ev->x -= a->x; ev->y -= a->y;
					handle_mouse_event_in_scroll_area(a, ev);
					return 1;
				}
				l = l->next;
			}
			break;
		case	YZ_MOUSE_UP:
			while(l)
			{
				a = l->e;
				if(a->in_drag)
				{
					ev->x -= a->x; ev->y -= a->y;
					handle_mouse_event_in_scroll_area(a, ev);
					return 1;
				}
				l = l->next;
			}
			break;
		case	YZ_MOUSE_DRAG:
			while(l)
			{
				a = l->e;
				if(a->in_drag)
				{
					ev->x -= a->x; ev->y -= a->y;
					ev->x2 -= a->x; ev->y2 -= a->y;
					handle_mouse_event_in_scroll_area(a, ev);
					return 1;
				}
				l = l->next;
			}
			break;
	}
	return 0;
}

static int	handle_mouse_event_in_scroll_area(scrollarea_t	*a, event_t	*ev)
{
	ev->x -= a->ox;
	ev->y -= a->oy;
	if(ev->type == YZ_MOUSE_DOWN)a->in_drag = 1;
	if(ev->type == YZ_MOUSE_UP)a->in_drag = 0;
	if(ev->type == YZ_MOUSE_DRAG) { ev->x2 -= a->ox; ev->y2 -= a->oy; }
	if(a->event)
	{
		yzPushClipRect(a->x, a->y, a->x+a->dx-1, a->y+a->dy-1);
		yzPushOrigin(a->x + a->ox, a->y + a->oy);
		a->event(a->data, *ev);
		// careful; event() may have destroyed this scrollarea!
		yzPopOrigin();
		yzPopClipRect();
	}
	return 1;
}
