#include	<stdlib.h>
#include	<string.h>

#include	<yz/window.h>
#include	<yz/draw.h>

#define	YZ_UITK_IMPLEMENTOR
#include	"uitk.h"
#include	"scrollbar.h"

struct scrollbar_t
{
	int	type;
	int	x, y;
	int	dx, dy;
	int	id;
	scrollarea_t	*area;

	int	in_drag, drag_ry;
	float	drag_now;

	int	old_thumb_y1, old_thumb_y2;
	int	old_scroll_at;
	
	float	min, max, now, page_size;
	int	axis, active;
};

static int		point_in_coords(int	x, int	y, int	x0, int	y0, int	x1, int	y1);
static int		handle_mouse_down_in_scroll_bar(scrollbar_t	*sb, int	rx, int	ry);
static int		handle_mouse_up_in_scroll_bar(scrollbar_t	*sb, int	rx, int	ry);
static int		handle_mouse_drag_in_scroll_bar(scrollbar_t	*sb, int	rx, int	ry,
														int	rx2, int	ry2);

#define	TYPE_NAME	"scroll-bar"

static int						this_type_filter(event_t	*ev);
static int						thisTypeNum = 0;
static ui_element_type_info		typeInfo = { TYPE_NAME, 0, 0, this_type_filter };

static int register_this_type()
{
	thisTypeNum = yzUIRegisterTypeInfo(&typeInfo);
	if(thisTypeNum==0)return -1;
	return 0;
}

#define	CHECK_REG()	if(register_this_type())return 0;

scrollbar_t	*yzAddScrollBar(int	x, int	y, int	dx, int	dy, int	id)
{
	scrollbar_t	*sbt;
	
	CHECK_REG();
	sbt = (scrollbar_t*)malloc(sizeof(scrollbar_t));
	sbt->type = thisTypeNum;
	sbt->x = x; sbt->y = y;
	sbt->dx = dx; sbt->dy = dy;
	sbt->id = id;
	sbt->in_drag = 0;
	sbt->min = 0;
	sbt->max = 1;
	sbt->now = 0; sbt->page_size = 1;
	sbt->old_thumb_y1 = 0;
	sbt->old_thumb_y2 = (dy>dx)?(dy-1):(dx-1);
	sbt->old_scroll_at = 0;
	sbt->area = 0;
	if(dx>dy)sbt->axis = 1;
	else sbt->axis = 0;
	sbt->active = 0;
	yzUIAddToList(sbt);
	yzDrawScrollBar(sbt);
	return sbt;
}

void	yzResizeScrollBar(scrollbar_t	*sbt, int	x, int	y, int	dx, int	dy)
{
	sbt->x = x; sbt->y = y;
	sbt->dx = dx; sbt->dy = dy;
	sbt->old_thumb_y1 = 0;
	sbt->old_thumb_y2 = (dy>dx)?(dy-1):(dx-1);
	if(dx>dy)sbt->axis = 1;
	else sbt->axis = 0;
}

void	yzRemoveScrollBar(scrollbar_t	*sbt)
{
	if(!sbt)return;
	yzUIRemoveFromList(sbt);
	free(sbt);
}

void		yzSetScrollBarMax(scrollbar_t	*sb, float	max)
{
	sb->max = max;
	if(sb->max-sb->min <= sb->page_size)
		sb->active = 0;
	else sb->active = 1;
	if(!yzSetScrollBarPosition(sb, sb->now))	// snap to legal values
		yzDrawScrollBar(sb);
}

void		yzSetScrollBarMin(scrollbar_t	*sb, float	min)
{
	sb->min = min;
	if(sb->max-sb->min <= sb->page_size)
		sb->active = 0;
	else sb->active = 1;
	if(!yzSetScrollBarPosition(sb, sb->now))	// snap to legal values
		yzDrawScrollBar(sb);
}

float			yzGetScrollBarPosition(scrollbar_t	*sb) { return sb->now; }

int			yzSetScrollBarPosition(scrollbar_t	*sb, float	now)
{
	float	delta;
	int		delta_i;

	if(now + sb->page_size > sb->max)now = sb->max - sb->page_size;
	if(now < sb->min)now = sb->min;
	delta = now - sb->now;
	if(delta)
	{
		sb->now = now;
		delta_i = sb->old_scroll_at - (int)now;

		yzDrawScrollBar(sb);
		if(sb->area && delta_i)
		{
			if(sb->axis)effect_scroll(sb->area, delta_i, 0);
			else effect_scroll(sb->area, 0, delta_i);
			sb->old_scroll_at = (int)now;
		}
		return 1;
	}
	else return 0;
}

void		yzSetScrollBarPageSize(scrollbar_t	*sb, float	ps)
{
	sb->page_size = ps;
	if(sb->max-sb->min <= sb->page_size)
		sb->active = 0;
	else sb->active = 1;
	if(!yzSetScrollBarPosition(sb, sb->now))	// snap to legal values
		yzDrawScrollBar(sb);
}

void		yzSetScrollBarArea(scrollbar_t	*sb, scrollarea_t	*a)
{
	sb->area = a;
	yzDrawScrollArea(a);
}

static int		this_type_filter(event_t	*ev)
{
	ui_element_list	*l;
	scrollbar_t		*sb;
	
	l = yzUIGetList(TYPE_NAME);
	if(!l)return 0;
	switch(ev->type)
	{
		case	YZ_MOUSE_DOWN:
			while(l)
			{
				sb = l->e;
				if(sb->active && point_in_coords(ev->x, ev->y, sb->x, sb->y,
					sb->x+sb->dx-1, sb->y+sb->dy-1))
					if(handle_mouse_down_in_scroll_bar(sb, ev->x - sb->x, ev->y - sb->y))
						return 1;
				l = l->next;
			}
			break;
		case	YZ_MOUSE_UP:
			while(l)
			{
				sb = l->e;
				if(sb->active && sb->in_drag)
					if(handle_mouse_up_in_scroll_bar(sb, ev->x - sb->x, ev->y - sb->y))
						return 1;
				l = l->next;
			}
			break;
		case	YZ_MOUSE_DRAG:
			while(l)
			{
				sb = l->e;
				if(sb->active && sb->in_drag)
					if(handle_mouse_drag_in_scroll_bar(sb, ev->x - sb->x, ev->y - sb->y,
						ev->x2 - sb->x, ev->y2 - sb->y))
						return 1;
				l = l->next;
			}
			break;
	}
	return 0;
}

static int	point_in_coords(int	x, int	y, int	x0, int	y0, int	x1, int	y1)
{
	if(x>=x0 && x<=x1 && y>=y0 && y<=y1)return 1;
	return 0;
}

draw_arrow_fore(int	x1, int	y1, int	x2, int	y2, int	dx, int	dy)
{
	float	x=(x1+x2)/2, y=(y1+y2)/2;

	yzPenColor(45000, 45000, 50000);
	yzRect(x1,y1,x2,y2);
	yzPenColor(0,0,0);
	yzOutlineRect(x1,y1,x2,y2);
	yzPenColor(20000, 20000, 30000);
	yzPolygon(3,	(int)(x-2*dx-4*dy+0.5), (int)(y+4*dx-2*dy+0.5),
					(int)(x-2*dx+4*dy+0.5), (int)(y-4*dx-2*dy+0.5),
					(int)(x+5*dx+0.5),		(int)(y+5*dy+0.5));
}

draw_arrow_aft(int	x1, int	y1, int	x2, int	y2, int	dx, int	dy)
{
	float	x=(x1+x2)/2, y=(y1+y2)/2;

	yzPenColor(45000, 45000, 50000);
	yzRect(x1,y1,x2,y2);
	yzPenColor(0,0,0);
	yzOutlineRect(x1,y1,x2,y2);
	yzPenColor(20000, 20000, 30000);
	yzPolygon(3,	(int)(x-4*dx-4*dy+0.5), (int)(y+4*dx-4*dy+0.5),
					(int)(x-4*dx+4*dy+0.5), (int)(y-4*dx-4*dy+0.5),
					(int)(x+4*dx+0.5),		(int)(y+4*dy+0.5));
}

#define	ARROW_SIZE	15

void	yzDrawScrollBar(scrollbar_t	*sb)
{
	int	thumb_top, thumb_bot, bar_top, bar_bot, bar_d;

	if(!sb->active)
	{
		// don't need buffer modes for this
		yzBufferMode(YZ_BUFFER_BOTH);
		yzPenColor(55000,55000,55000);
		yzRect(sb->x+1, sb->y+1, sb->x+sb->dx-2, sb->y+sb->dy-2);
		yzPenColor(0,0,0);
		yzOutlineRect(sb->x, sb->y, sb->x+sb->dx-1, sb->y+sb->dy-1);
		sb->old_thumb_y1 = 0;
		sb->old_thumb_y2 = (sb->axis==1)?(sb->dx):(sb->dy);
		return;
	}

	yzBufferMode(YZ_BUFFER_BACK);

	// erase the old thumb
	if(sb->axis==1)	// horizontal
	{
		bar_top = sb->x + ARROW_SIZE;
		bar_bot = sb->x+sb->dx-1 - ARROW_SIZE;
		bar_d = sb->dx - ARROW_SIZE*2;
		yzPushClipRect(bar_top+sb->old_thumb_y1, sb->y, bar_top+sb->old_thumb_y2, sb->y+sb->dy-1);
		yzPenColor(50000, 50000, 55000);
		yzRect(bar_top, sb->y, bar_bot, sb->y+sb->dy-1);
		yzPenColor(0, 0, 0);
		yzOutlineRect(sb->x, sb->y, sb->x+sb->dx-1, sb->y+sb->dy-1);
		yzPopClipRect();
		thumb_top = (bar_d-1) * (sb->now / (sb->max - sb->min));
		thumb_bot = (bar_d-1) * ((sb->now+sb->page_size) / (sb->max - sb->min));
	}
	else	// vertical
	{
		bar_top = sb->y + ARROW_SIZE;
		bar_bot = sb->y+sb->dy-1 - ARROW_SIZE;
		bar_d = sb->dy - ARROW_SIZE*2;
		yzPushClipRect(sb->x, bar_top+sb->old_thumb_y1, sb->x+sb->dx-1, bar_top+sb->old_thumb_y2);
		yzPenColor(50000, 50000, 55000);
		yzRect(sb->x, bar_top, sb->x+sb->dx-1, bar_bot);
		yzPenColor(0, 0, 0);
		//printf("outlining %d %d %d %d\n",
					//sb->x, sb->y, sb->x+sb->dx-1, sb->y+sb->dy);
		yzOutlineRect(sb->x, sb->y, sb->x+sb->dx-1, sb->y+sb->dy-1);
		yzPopClipRect();
		thumb_top = (bar_d-1) * (sb->now / (sb->max - sb->min));
		thumb_bot = (bar_d-1) * ((sb->now+sb->page_size) / (sb->max - sb->min));
	}
	
	if(sb->in_drag)yzPenColor(35000, 45000, 55000);
	else yzPenColor(40000, 40000, 45000);

	// draw the new thumb and arrows
	if(sb->axis==1)	// horizontal
	{
		yzRect(bar_top + thumb_top, sb->y + 1, bar_top + thumb_bot, sb->y+sb->dy-2);
		yzPenColor(0, 0, 0);
		yzOutlineRect(bar_top+thumb_top, sb->y + 1, bar_top+thumb_bot, sb->y + sb->dy-2);
		draw_arrow_aft(sb->x, sb->y, bar_top-1, sb->y+sb->dy-1, -1, 0);
		draw_arrow_fore(bar_bot+1, sb->y, sb->x+sb->dx-1, sb->y+sb->dy-1, 1, 0);
	}
	else	// vertical
	{
		yzRect(sb->x+1, bar_top + thumb_top, sb->x+sb->dx-2, bar_top + thumb_bot);
		yzPenColor(0, 0, 0);
		yzOutlineRect(sb->x+1, bar_top + thumb_top, sb->x+sb->dx-2, bar_top + thumb_bot);
		draw_arrow_aft(sb->x, sb->y, sb->x+sb->dx-1, bar_top-1, 0, -1);
		draw_arrow_fore(sb->x, bar_bot+1, sb->x+sb->dx-1, sb->y+sb->dy-1, 0, 1);
	}
	yzBufferMode(YZ_BUFFER_BOTH);
	yzPushClipRect(sb->x, sb->y, sb->x+sb->dx-1, sb->y+sb->dy-1);
	yzUpdateBuffer(YZ_BUFFER_FRONT);
	yzPopClipRect();

	sb->old_thumb_y1 = thumb_top;
	sb->old_thumb_y2 = thumb_bot;
}

static int		handle_mouse_down_in_scroll_bar(scrollbar_t	*sb, int	rx, int	ry)
{
	int	thumb_top, thumb_bot;
	int	dist, mval;
	int	repeat = 0, repeat_ms = 200;
	//float	perpix;

	if(sb->axis){dist=sb->dx;mval = rx;}
	else {dist = sb->dy;mval = ry;}
	dist -= ARROW_SIZE*2;
	//perpix = ((float)(sb->max-sb->min)) / dist;

	do
	{
		thumb_top = ARROW_SIZE + dist * (sb->now / (sb->max - sb->min));
		thumb_bot = ARROW_SIZE + dist * ((sb->now+sb->page_size) / (sb->max - sb->min));
		if(mval<thumb_top)
		{
			// line up
			if(mval < ARROW_SIZE)
				yzSetScrollBarPosition(sb, sb->now - 15);
			// page up
			else yzSetScrollBarPosition(sb, sb->now - sb->page_size);
			repeat = 1;
		}
		else if(mval>thumb_bot)
		{
			// line down
			if(mval > dist+ARROW_SIZE)
				yzSetScrollBarPosition(sb, sb->now + 15);
			// page down
			else yzSetScrollBarPosition(sb, sb->now + sb->page_size);
			repeat = 1;
		}
		else
		{
			// thumb
			sb->in_drag = 1;
			sb->drag_ry = mval;
			sb->drag_now = sb->now;
			yzDrawScrollBar(sb);
		}
		if(repeat)
		{
			event_t	ev = yzGetUnfilteredEvent(repeat_ms);
			if(ev.type == YZ_MOUSE_UP)
				repeat = 0;
			repeat_ms = 50;
		}
	} while(repeat);
	return 1;
}

static int		handle_mouse_up_in_scroll_bar(scrollbar_t	*sb, int	rx, int	ry)
{
	if(sb->in_drag)
	{
		sb->in_drag = 0;
		yzDrawScrollBar(sb);
		return 1;
	}
	else return 0;
}

static int		handle_mouse_drag_in_scroll_bar(scrollbar_t	*sb, int	rx, int	ry,
														int	rx2, int	ry2)
{
	if(sb->in_drag)
	{
		if(sb->axis)yzSetScrollBarPosition(sb, sb->drag_now +
				( (float)(rx2 - sb->drag_ry) ) / (sb->dx-ARROW_SIZE*2) * (sb->max - sb->min) );
		else yzSetScrollBarPosition(sb, sb->drag_now +
				( (float)(ry2 - sb->drag_ry) ) / (sb->dy-ARROW_SIZE*2) * (sb->max - sb->min) );
		return 1;
	}
	else return 0;
}
